/* Copyright (C) 2007-2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Maksym Zyzak, Valentina Akishina, Igor Kulakov [committer], Sergei Zharko */

#pragma once  // include this header only once per compilation unit

#include "CaBranch.h"
#include "CaCloneMerger.h"
#include "CaConstants.h"
#include "CaField.h"
#include "CaGrid.h"
#include "CaGridEntry.h"
#include "CaHit.h"
#include "CaInputData.h"
#include "CaMeasurementXy.h"
#include "CaParameters.h"
#include "CaStation.h"
#include "CaTimer.h"
#include "CaTrack.h"
#include "CaTrackExtender.h"
#include "CaTrackFinder.h"
#include "CaTrackFinderWindow.h"
#include "CaTrackFitter.h"
#include "CaTrackParam.h"
#include "CaTrackingMonitor.h"
#include "CaTriplet.h"
#include "CaVector.h"

#include <array>
#include <iomanip>
#include <iostream>
#include <limits>
#include <map>

namespace cbm::algo::ca
{
  class TripletConstructor;


  namespace
  {
    using namespace cbm::algo;  // to keep ca:: prefices in the code
  }

  //namespace
  //{
  //  using cbm::algo::ca::Track;      // TMP
  //  using cbm::algo::ca::Vector;     // TMP
  //  using cbm::algo::ca::Iteration;  // TMP
  //}  // namespace

  // *******************************
  // ** Types definition (global) **
  // *******************************

  using CaStationsArray_t = std::array<ca::StationV, constants::size::MaxNstations>;
  using CaMaterialArray_t = std::array<ca::MaterialMap, constants::size::MaxNstations>;
  using Tindex            = int;  // TODO: Replace with ca::HitIndex_t, if suitable

  struct CaHitTimeInfo {
    fscal fEventTimeMin{-std::numeric_limits<fscal>::max() / 2.};
    fscal fEventTimeMax{std::numeric_limits<fscal>::max() / 2.};
    fscal fMaxTimeBeforeHit{0.};  //< max event time for hits [0 .. hit] in the station hit array
    fscal fMinTimeAfterHit{0.};   //< min event time for hits [hit ... ] in the station hit array
  };

  /// Main class of CA track finder algorithm
  ///
  class Framework {
   public:
    // *************************
    // ** Friend classes list **
    // *************************

    friend class ca::TripletConstructor;
    friend class ca::TrackExtender;
    friend class ca::TrackFitter;
    friend class ca::TrackFinderWindow;
    friend class ca::TrackFinder;

    ///---------------------------
    /// Internal enumerations

    enum TrackingMode
    {
      kSts,
      kGlobal,
      kMcbm
    };

    // **********************************
    // ** Member functions declaration **
    // **********************************

    // ** Constructors and destructor

    /// Constructor
    Framework();

    /// Copy constructor
    Framework(const Framework&) = delete;

    /// Move constructor
    Framework(Framework&&) = delete;

    /// Copy assignment operator
    Framework& operator=(const Framework&) = delete;

    /// Move assignment operator
    Framework& operator=(Framework&&) = delete;

    /// Destructor
    ~Framework() = default;


    // ** Functions, which pack and unpack indexes of station and triplet **

    // TODO: move to ca::Triplet class (S.Zharko)
    /// Packs station and triplet indices to an unique triplet ID
    /// \param  iStation  Index of station in the active stations array
    /// \param  iTriplet  Index of triplet
    static unsigned int PackTripletId(unsigned int iStation, unsigned int iTriplet);

    /// Unpacks the triplet ID to its station index
    /// \param  id  Unique triplet ID
    static unsigned int TripletId2Station(unsigned int id);

    /// Unpacks the triplet ID to its triplet index
    /// \param  id  Unique triplet ID
    static unsigned int TripletId2Triplet(unsigned int id);


    /// Sets Framework parameters object
    /// \param other - reference to the Parameters object
    void SetParameters(const Parameters& other) { fParameters = other; }
    // TODO: remove it (S.Zharko)

    /// Gets a pointer to the Framework parameters object
    const Parameters& GetParameters() const { return fParameters; }

    /// Receives input data
    void ReceiveInputData(InputData&& inputData);

    /// Receives tracking parameters
    void ReceiveParameters(Parameters&& parameters);

    /// Gets pointer to input data object for external access
    const InputData& GetInputData() const { return fInputData; }

    int PackIndex(const int& a, const int& b, const int& c);

    int UnPackIndex(const int& i, int& a, int& b, int& c);

    /// \brief Sets a default particle mass for the track fit
    ///
    /// The function is used during the reconstruction in order to estimate the multiple scattering and energy loss
    /// \param mass Default particle mass
    void SetDefaultParticleMass(float mass) { fDefaultMass = mass; }

    /// Gets default particle mass
    /// \return particle mass
    float GetDefaultParticleMass() const { return fDefaultMass; }

    /// Gets default particle mass squared
    /// \return particle mass squared
    float GetDefaultParticleMass2() const { return fDefaultMass * fDefaultMass; }


    /*********************************************************************************************/ /**
   *                             ------  FUNCTIONAL PART ------
   ************************************************************************************************/


    void Init(const TrackingMode mode);

    void Finish();

    void PrintHits();

    float GetMaxInvMom() const { return fMaxInvMom[0]; }

    /// \brief Gets monitor data
    const TrackingMonitorData& GetMonitorData() const { return fMonitorData; }

    /// \brief Sets monitor data
    void SetMonitorData(const TrackingMonitorData& monitorData) { fMonitorData = monitorData; }

   public:
    /// Gets number of stations before the pipe (MVD stations in CBM)
    // int GetNstationsBeforePipe() const { return fNstationsBeforePipe; }

    /// Gets number of stations situated in field region (MVD + STS in CBM)
    int GetNfieldStations() const { return fNfieldStations; }

    /// Get mc track ID for a hit (debug tool)
    int GetMcTrackIdForCaHit(int iHit) const;

    /// Get mc track ID for a hit (debug tool)
    int GetMcTrackIdForWindowHit(int iGridHit) const;

    // const CbmL1MCTrack* GetMcTrackForWindowHit(int iHit) const;

   private:
    int fNstationsBeforePipe{0};                    ///< number of stations before pipe (MVD stations in CBM)
    int fNfieldStations{0};                         ///< number of stations in the field region
    float fDefaultMass{constants::phys::MuonMass};  ///< mass of the propagated particle [GeV/c2]


    // ***************************
    // ** Member variables list **
    // ***************************

    Parameters fParameters;  ///< Object of Framework parameters class
    InputData fInputData;    ///< Tracking input data

    Vector<unsigned char> fvHitKeyFlags{
      "Framework::fvHitKeyFlags"};  ///< List of key flags: has been this hit or cluster already used

    TrackingMonitorData fMonitorData{};  ///< Tracking monitor data (statistics per call)

   public:
    Vector<CaHitTimeInfo> fHitTimeInfo;

    ca::Grid vGrid[constants::size::MaxNstations];  ///<

    /// ----- Data used during track finding -----
    ///

    ///\brief hits of the current time window
    /// it is a portion of fInputData to process in the current time window
    /// hit.Id is replaced by the hit index in fInputData
    Vector<ca::Hit> fWindowHits{"Framework::fWindowHits"};

    Vector<unsigned char> fIsWindowHitSuppressed{"Framework::fIsWindowHitSuppressed"};

    ///\brief first index of the station hits in the time window
    ca::HitIndex_t fStationHitsStartIndex[constants::size::MaxNstations + 1]{0};

    ///\brief number of station hits in the time window
    ca::HitIndex_t fStationNhits[constants::size::MaxNstations + 1]{0};

    /// ----------

    double fCaRecoTime{0.};  // time of the track finder + fitter

    Vector<Track> fRecoTracks{"Framework::fRecoTracks"};       ///< reconstructed tracks
    Vector<ca::HitIndex_t> fRecoHits{"Framework::fRecoHits"};  ///< packed hits of reconstructed tracks

    Vector<Track> fSliceRecoTracks{"Framework::fSliceRecoTracks"};       ///< reconstructed tracks in sub-timeslice
    Vector<ca::HitIndex_t> fSliceRecoHits{"Framework::fSliceRecoHits"};  ///< packed hits of reconstructed tracks

    /// Created triplets vs station index
    Vector<ca::Triplet> fTriplets[constants::size::MaxNstations]{{"Framework::fTriplets"}};

    /// Track candidates created out of adjacent triplets before the final track selection.
    /// The candidates may share any amount of hits.
    Vector<ca::Branch> fTrackCandidates{"Framework::fTrackCandidates"};

    ///< indices of the sub-slice hits
    Vector<ca::HitIndex_t> fSliceHitIds[constants::size::MaxNstations]{"Framework::fSliceHitIds"};

    Vector<int> fHitKeyToTrack{"Framework::fHitKeyToTrack"};  // strip to track pointers

    TrackingMode fTrackingMode{kSts};

    fvec EventTime{0.f};
    fvec Err{0.f};

    /// --- data used during finding iterations
    unsigned int fCurrentIterationIndex{0};         ///< index of the corrent iteration, needed for debug output
    const Iteration* fpCurrentIteration = nullptr;  ///< pointer to the current CA track finder iteration

    Vector<int> fHitFirstTriplet{"Framework::fHitFirstTriplet"};  /// link hit -> first triplet { hit, *, *}
    Vector<int> fHitNtriplets{"Framework::fHitNtriplets"};        /// link hit ->n triplets { hit, *, *}


   public:
    ca::TrackExtender fTrackExtender{*this};          ///< Object of the track extender algorithm
    ca::CloneMerger fCloneMerger{*this};              ///< Object of  the clone merger algorithm
    ca::TrackFitter fTrackFitter{*this};              ///< Object of the track extender algorithm
    ca::TrackFinderWindow fTrackFinderWindow{*this};  ///< Object of the track finder algorithm for the time window
    ca::TrackFinder fTrackFinder{*this};  ///< Object of the track finder algorithm for the entire time slice

   private:
    /// =================================  DATA PART  =================================

    /// ----- Different parameters of CATrackFinder -----

    Tindex fFirstCAstation{0};  //first station used in CA

    // fNFindIterations - set number of interation for trackfinding
    // itetation of finding:

    int fNFindIterations = -1;  // TODO investigate kAllPrimJumpIter & kAllSecJumpIter

    float fTrackChi2Cut{10.f};
    float fTripletFinalChi2Cut{10.f};
    float fTripletChi2Cut{5.f};  // cut for selecting triplets before collecting tracks.per one DoF
    float fDoubletChi2Cut{5.f};
    float fTimeCut1{0.f};  // TODO: please, specify "1" and "2" (S.Zharko)
    float fTimeCut2{0.f};

    /// correction in order to take into account overlaping and iff z. if sort by y then it is max diff between same station's modules (~0.4cm)
    fvec fMaxDZ{constants::Undef<fvec>};

    /// parameters which are different for different iterations. Set in the begin of CA Track Finder

    float fPickGather{constants::Undef<fvec>};  ///< same for attaching additional hits to track
    float fTripletLinkChi2{
      constants::Undef<fvec>};                 ///< (dp2/dp_error2 < fTripletLinkChi2)  =>  triplets are neighbours
    fvec fMaxInvMom{constants::Undef<fvec>};   ///< max considered q/p for tracks
    fvec fMaxSlopePV{constants::Undef<fvec>};  ///< max slope (tx\ty) in prim vertex
    float fMaxSlope{constants::Undef<fvec>};   ///< max slope (tx\ty) in 3d hit position of a triplet

    fvec fTargX{constants::Undef<fvec>};  ///< target position x coordinate for the current iteration (modifiable)
    fvec fTargY{constants::Undef<fvec>};  ///< target position y coordinate for the current iteration (modifiable)
    fvec fTargZ{constants::Undef<fvec>};  ///< target position z coordinate for the current iteration (modifiable)

    bool fIsTargetField{false};  ///< is the magnetic field present at the target

    ca::FieldValueV fTargB _fvecalignment{};  // field in the target point (modifiable, do not touch!!)
    ca::MeasurementXy<fvec> TargetMeasurement _fvecalignment{};  // target constraint  [cm]

    int fGhostSuppression{1};  // NOTE: Should be equal to 0 in TRACKS_FROM_TRIPLETS mode!

  } _fvecalignment;


  // ********************************************
  // ** Inline member functions implementation **
  // ********************************************

  // ---------------------------------------------------------------------------------------------------------------------
  //
  [[gnu::always_inline]] inline unsigned int Framework::PackTripletId(unsigned int iStation, unsigned int iTriplet)
  {
#ifndef FAST_CODE
    assert(iStation < constants::size::MaxNstations);
    assert(iTriplet < constants::size::MaxNtriplets);
#endif
    constexpr unsigned int kMoveStation = constants::size::TripletBits;
    return (iStation << kMoveStation) + iTriplet;
  }

  // ---------------------------------------------------------------------------------------------------------------------
  //
  [[gnu::always_inline]] inline unsigned int Framework::TripletId2Station(unsigned int id)
  {
    constexpr unsigned int kMoveStation = constants::size::TripletBits;
    return id >> kMoveStation;
  }

  // ---------------------------------------------------------------------------------------------------------------------
  //
  [[gnu::always_inline]] inline unsigned int Framework::TripletId2Triplet(unsigned int id)
  {
    constexpr unsigned int kTripletMask = (1u << constants::size::TripletBits) - 1u;
    return id & kTripletMask;
  }

}  // namespace cbm::algo::ca
